"""Internal helpers."""

from __future__ import annotations
import argparse
import re
import shlex
import subprocess
import typing

if typing.TYPE_CHECKING:
    from typing import Dict, Any, Tuple, Mapping, Optional, Type
    from argparse import ArgumentParser, Namespace
    from .base import Section


class Switch(argparse.Action):
    """Inherited from argparse.Action, store True/False to a +/-arg.

    The :func:`switch_opt` function allows you to easily create a
    :class:`~loam.tools.ConfOpt` using this action.
    """

    def __call__(self, parser: ArgumentParser, namespace: Namespace,
                 values: Any, option_string: Optional[str] = None) -> None:
        """Set args attribute to True or False."""
        if option_string is None:
            raise ValueError("Switch action is not suitable for "
                             "positional arguments.")
        setattr(namespace, self.dest, bool('-+'.index(option_string[0])))


class SectionContext:
    """Context manager to locally change option values.

    It is reusable but not reentrant.

    Args:
        section (Section): configuration section to be managed.
        options (Mapping): mapping between option names and their values in the
            context.
    """

    def __init__(self, section: Section, options: Mapping[str, Any]):
        self._section = section
        self._options = options
        self._old_values: Dict[str, Any] = {}

    def __enter__(self) -> None:
        self._old_values = {
            opt: getattr(self._section, opt) for opt in self._options
        }
        self._section.update_from_dict_(self._options)

    def __exit__(self, e_type: Optional[Type[BaseException]], *_: Any) -> bool:
        self._section.update_from_dict_(self._old_values)
        return e_type is None


def zsh_version() -> Tuple[int, ...]:
    """Try to guess zsh version, return (0, 0) on failure."""
    try:
        out = subprocess.run(shlex.split('zsh --version'), check=True,
                             stdout=subprocess.PIPE).stdout
    except (FileNotFoundError, subprocess.CalledProcessError):
        return (0, 0)
    v_match = re.search(br'[0-9]+\.[0-9]+', out)
    return tuple(map(int, v_match.group(0).split(b'.'))) if v_match else (0, 0)



import toml


def test_create_config(conf, cfile):
    conf.to_file_(cfile)
    conf_dict = toml.load(str(cfile))
    assert conf_dict == {'sectionA': {'optA': 1, 'optC': 3, 'optBool': True},
                         'sectionB': {'optA': 4, 'optC': 6, 'optBool': False}}


def test_create_config_no_update(conf, cfile):
    conf.sectionA.optA = 42
    conf.default_().to_file_(cfile)
    conf_dict = toml.load(str(cfile))
    assert conf_dict == {'sectionA': {'optA': 1, 'optC': 3, 'optBool': True},
                         'sectionB': {'optA': 4, 'optC': 6, 'optBool': False}}


def test_create_config_update(conf, cfile):
    conf.sectionA.optA = 42
    conf.to_file_(cfile)
    conf_dict = toml.load(str(cfile))
    assert conf_dict == {'sectionA': {'optA': 42, 'optC': 3, 'optBool': True},
                         'sectionB': {'optA': 4, 'optC': 6, 'optBool': False}}



"""Parsers for your CLI arguments.

These functions can be used as `from_toml` in :attr:`~loam.base.Entry`.
"""

from __future__ import annotations
from typing import Union


def strict_slice_parser(arg: object) -> slice:
    """Parse a string into a slice.

    Note that this errors out on a single integer with no `:`.  If you
    want to treat a single integer as a slice from 0 to that value, see
    :func:`slice_parser`.  To treat a single integer as an integer, see
    :func:`slice_or_int`.
    """
    soi = slice_or_int_parser(arg)
    if isinstance(soi, int):
        raise ValueError(f"{arg} is an invalid slice")
    return soi


def slice_parser(arg: object) -> slice:
    """Parse a string into a slice.

    Note that this treats a single integer as a slice from 0 to that
    value.  To error out on a single integer, use :func:`strict_slice_parser`.
    To parse it as an integer, use :func:`slice_or_int_parser`.
    """
    soi = slice_or_int_parser(arg)
    if isinstance(soi, int):
        return slice(soi)
    return soi


def slice_or_int_parser(arg: object) -> Union[slice, int]:
    """Parse a string into a slice.

    Note that this treats a single integer as an integer value.  To error out
    on a single integer, use :func:`strict_slice_parser`.  To parse it as a
    slice, use :func:`slice_parser`.
    """
    if isinstance(arg, int):
        return arg
    if isinstance(arg, slice):
        return arg
    if not isinstance(arg, str):
        raise TypeError("arg should be an int, slice, or str")
    if ':' in arg:
        idxs = arg.split(':')
        if len(idxs) > 3:
            raise ValueError(f'{arg} is an invalid slice')
        slice_parts = [int(idxs[0]) if idxs[0] else None,
                       int(idxs[1]) if idxs[1] else None]
        if len(idxs) == 3:
            slice_parts.append(int(idxs[2]) if idxs[2] else None)
        else:
            slice_parts.append(None)
        return slice(*slice_parts)
    return int(arg)



import pytest
from loam import parsers


def test_slice_or_int_parser():
    assert parsers.slice_or_int_parser(28) == 28
    assert parsers.slice_or_int_parser(slice(2, 28, 5)) == slice(2, 28, 5)
    assert parsers.slice_or_int_parser("42") == 42
    assert parsers.slice_or_int_parser(":3") == slice(3)
    assert parsers.slice_or_int_parser("1:3") == slice(1, 3)
    assert parsers.slice_or_int_parser("1:") == slice(1, None)
    assert parsers.slice_or_int_parser("23:54:2") == slice(23, 54, 2)
    assert parsers.slice_or_int_parser("::5") == slice(None, None, 5)
    with pytest.raises(ValueError):
        parsers.slice_or_int_parser("1:2:3:4")
    with pytest.raises(TypeError):
        parsers.slice_or_int_parser(object())


def test_strict_slice_parser():
    with pytest.raises(ValueError):
        parsers.strict_slice_parser(28)
    assert parsers.strict_slice_parser(slice(2, 28, 5)) == slice(2, 28, 5)
    with pytest.raises(ValueError):
        parsers.strict_slice_parser("42")
    assert parsers.strict_slice_parser(":3") == slice(3)
    assert parsers.strict_slice_parser("1:3") == slice(1, 3)
    assert parsers.strict_slice_parser("1:") == slice(1, None)
    assert parsers.strict_slice_parser("23:54:2") == slice(23, 54, 2)
    assert parsers.strict_slice_parser("::5") == slice(None, None, 5)
    with pytest.raises(TypeError):
        parsers.strict_slice_parser(object())


def test_slice_parser():
    assert parsers.slice_parser(28) == slice(28)
    assert parsers.slice_parser(slice(2, 28, 5)) == slice(2, 28, 5)
    assert parsers.slice_parser("42") == slice(42)
    assert parsers.slice_parser(":3") == slice(3)
    assert parsers.slice_parser("1:3") == slice(1, 3)
    assert parsers.slice_parser("1:") == slice(1, None)
    assert parsers.slice_parser("23:54:2") == slice(23, 54, 2)
    assert parsers.slice_parser("::5") == slice(None, None, 5)
    with pytest.raises(TypeError):
        parsers.slice_parser(object())



from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

import pytest

from loam.base import entry, Section, ConfigBase


class MyMut:
    def __init__(self, inner_list):
        if not isinstance(inner_list, list):
            raise TypeError
        self.inner_list = inner_list

    @staticmethod
    def from_toml(s: object) -> MyMut:
        if isinstance(s, str):
            return MyMut(list(map(float, s.split(","))))
        else:
            raise TypeError


def test_with_val():
    @dataclass
    class MySection(Section):
        some_n: int = entry(val=42)

    sec = MySection()
    assert sec.some_n == 42


def test_two_vals_fail():
    with pytest.raises(ValueError):
        entry(val=5, val_factory=lambda: 5)


def test_cast_and_set_type_hint(section_a):
    assert section_a.some_n == 42
    assert section_a.some_str == "foo"
    section_a.cast_and_set_("some_n", "5")
    assert section_a.some_n == 5
    section_a.cast_and_set_("some_str", "bar")
    assert section_a.some_str == "bar"


def test_context(section_a):
    with section_a.context_(some_n=5, some_str="bar"):
        assert section_a.some_n == 5
        assert section_a.some_str == "bar"
    assert section_a.some_n == 42
    assert section_a.some_str == "foo"


def test_context_cast(section_b):
    with section_b.context_(some_path="my/path"):
        assert section_b.some_path == Path("my/path")
    assert section_b.some_path == Path()


def test_cast_mutable_protected():
    @dataclass
    class MySection(Section):
        some_mut: MyMut = entry(val_toml="4.5,3.8", from_toml=MyMut.from_toml)

    MySection().some_mut.inner_list.append(5.6)
    assert MySection().some_mut.inner_list == [4.5, 3.8]


def test_type_hint_not_a_class():
    @dataclass
    class MySection(Section):
        maybe_n: Optional[int] = entry(val_factory=lambda: None, from_toml=int)
    assert MySection().maybe_n is None
    assert MySection("42").maybe_n == "42"


def test_with_obj_no_from_toml():
    with pytest.raises(ValueError):
        entry(val_toml="5")


def test_init_wrong_type():
    @dataclass
    class MySection(Section):
        some_n: int = 42
    with pytest.raises(TypeError):
        MySection("bla")


def test_missing_from_toml():
    @dataclass
    class MySection(Section):
        my_mut: MyMut = entry(val_factory=lambda: MyMut([4.5]))
    sec = MySection()
    assert sec.my_mut.inner_list == [4.5]
    with pytest.raises(TypeError):
        sec.cast_and_set_("my_mut", "4.5,3.8")


def test_config_default(my_config):
    assert my_config.section_a.some_n == 42
    assert my_config.section_b.some_path == Path()
    assert my_config.section_a.some_str == "foo"
    assert my_config.section_b.some_str == "bar"


def test_to_from_toml(my_config, cfile):
    my_config.section_a.some_n = 5
    my_config.section_b.some_path = Path("foo/bar")
    my_config.to_file_(cfile)
    new_config = my_config.default_()
    new_config.update_from_file_(cfile)
    assert my_config == new_config


def test_to_toml_not_in_file(my_config, cfile):
    my_config.section_b.some_str = "ignored"
    my_config.to_file_(cfile)
    content = cfile.read_text()
    assert "ignored" not in content
    assert "section_not_in_file" not in content


def test_from_toml_not_in_file(my_config, cfile):
    cfile.write_text('[section_b]\nsome_str="ignored"\n')
    my_config.default_().update_from_file_(cfile)
    assert my_config.section_b.some_str == "bar"


def test_to_file_exist_ok(my_config, cfile):
    my_config.to_file_(cfile)
    with pytest.raises(RuntimeError):
        my_config.to_file_(cfile, exist_ok=False)
    my_config.to_file_(cfile)


def test_config_with_not_section():
    @dataclass
    class MyConfig(ConfigBase):
        dummy: int = 5
    with pytest.raises(TypeError):
        MyConfig.default_()


def test_update_opt(conf):
    conf.sectionA.update_from_dict_({'optA': 42, 'optC': 43})
    assert conf.sectionA.optA == 42 and conf.sectionA.optC == 43


def test_update_section(conf):
    conf.update_from_dict_(
        {'sectionA': {'optA': 42}, 'sectionB': {'optA': 43}})
    assert conf.sectionA.optA == 42 and conf.sectionB.optA == 43



"""Exceptions raised by loam."""


class LoamError(Exception):
    """Base class for exceptions raised by loam."""

    pass


class LoamWarning(UserWarning):
    """Warning category for warnings issued by loam."""

    pass


class SubcmdError(LoamError):
    """Raised when an invalid Subcmd name is requested.

    Args:
        option: invalid subcommand name.

    Attributes:
        option: invalid subcommand name.
    """

    def __init__(self, option: str):
        self.option = option
        super().__init__(f'invalid subcommand name: {option}')
